from numpy.testing import assert_almost_equal, assert_equal
from pylj import forcefields
import unittest


class TestForcefields(unittest.TestCase):
    def test_lennard_jones(self):
        a = forcefields.lennard_jones([1.0, 1.0])
        assert_almost_equal(a.energy(2.0), -0.015380859)
        assert_almost_equal(a.force(2.0), -0.045410156)
        a.mixing([2.0, 3.0])
        assert_almost_equal([a.a, a.b], [1.4239324, 1.7379922])
        b = forcefields.lennard_jones([1.0, 1.0])
        assert_almost_equal(b.energy([2.0, 4.5],), [-0.015380859, -0.00012041])
        assert_almost_equal(b.force([2.0, 4.0]), [-0.045410156, -3.66032124e-04])
        c = forcefields.lennard_jones([0.5, 3.5])
        assert_almost_equal(c.energy([2.0, 4.5],), [-0.05456543, -0.00042149])
        assert_almost_equal(c.force([2.0, 4.5],), [-0.1633301, -0.000562])
        d = forcefields.lennard_jones( [5.0, 3.5])
        assert_almost_equal(d.energy([1.0, 1.5, 20.0]), [1.50000000, -0.268733500, -5.46874988e-08])
        assert_almost_equal(d.force([1.0, 1.5, 20.0]), [ 3.9000000e+01, -9.2078707e-01, -1.6406249e-08])
        e = forcefields.lennard_jones([100.0, 300.0])
        assert_almost_equal(e.energy([100.0, 200.0, 500.0]), [0, 0, 0])
        assert_almost_equal(e.force([100.0, 200.0, 500.0]), [0, 0, 0])
        with self.assertRaises(IndexError):
            f = forcefields.lennard_jones([1.0, 1.0, 1.0])

    def test_lennard_jones_sigma_epsilon(self):
        a = forcefields.lennard_jones_sigma_epsilon([1.0, 0.25])
        assert_almost_equal(a.energy(2.0), -0.015380859)
        assert_almost_equal(a.force(2.0), -0.045410156)
        b = forcefields.lennard_jones_sigma_epsilon([1.0, 0.25])
        assert_almost_equal(b.energy([2.0, 1.0]), [-0.015380859, 0])
        assert_almost_equal(b.force([2.0, 4.0]), [-0.0454102, -0.000366])
        c = forcefields.lennard_jones_sigma_epsilon([0.5, 0.75])
        assert_almost_equal(c.energy([2.0, 1.0, 1.5]), [-0.0007322, -0.0461425, -0.0041095])
        assert_almost_equal(c.force([2.0, 1.0, 1.5]), [-0.0021962, -0.2724609, -0.0164157])
        d = forcefields.lennard_jones_sigma_epsilon([5e-10, 9e-9])
        assert_almost_equal(d.energy([400.0, 500.0, 600.0]), [0, 0, 0])
        assert_almost_equal(d.force([400.0, 500.0, 600.0]), [0, 0, 0])
        e = forcefields.lennard_jones_sigma_epsilon([1.0, 1.0])
        e.mixing([4.0, 4.0])
        assert_almost_equal([e.sigma, e.epsilon], [2.5, 2.0])
        with self.assertRaises(IndexError):
            f = forcefields.lennard_jones_sigma_epsilon([1.0, 1.0, 1.0])

    def test_buckingham(self):
        a = forcefields.buckingham([1.0, 1.0, 1.0])
        assert_almost_equal(a.energy(2.0), 0.1197103832)
        assert_almost_equal(a.force(2.0), 0.08846028324)
        a.mixing([4.0, 4.0, 4.0])
        assert_almost_equal([a.a, a.b, a.c], [2.0, 2.0, 2.0])
        b = forcefields.buckingham([1.0, 1.0, 1.0])
        assert_almost_equal(b.energy([2.0]), 0.1197103832)
        assert_almost_equal(b.force([2.0]), 0.08846028324)
        c = forcefields.buckingham([1.0, 1.0, 1.0])
        assert_almost_equal(c.energy([2.0, 4.0]), [0.1197103832, 0.0180715])
        assert_almost_equal(c.force([2.0, 4.0]), [0.0884603, 0.0179494])
        d = forcefields.buckingham([0.01, 0.01, 0.01])
        assert_almost_equal(d.energy([2.0, 4.0, 5.0]), [0.0096457, 0.0096055, 0.0095117])
        assert_almost_equal(d.force([2.0, 4.0, 5.0]), [-3.7073013e-04,  9.2416835e-05,  9.4354942e-05])
        with self.assertRaises(IndexError):
            e = forcefields.buckingham([1.0, 1.0])

    def test_square_well(self):
        a = forcefields.square_well([1.0, 1.5, 2.0])
        assert_equal(a.energy(2.0), -1.0)
        with self.assertRaises(ValueError):
            a.force()
        b = forcefields.square_well([1.0, 2.0, 1.25])
        assert_equal(b.energy(0.5), float('inf'))
        c = forcefields.square_well([0.5, 1.5, 1.25])
        assert_equal(c.energy(3.0), 0)
        d = forcefields.square_well([1.0, 1.5, 2.0])
        assert_equal(d.energy([2.0, 0.5]), [-1.0, float('inf')])
        e = forcefields.square_well([1.0, 1.5, 1.25])
        assert_equal(e.energy([3.0, 3.0, 0.25]), [0, 0, float('inf')])
        f = forcefields.square_well([1.0, 1.5, 1.25], max_val=5000)
        assert_equal(f.energy([3.0, 3.0, 0.25]), [0, 0, 5000])
        with self.assertRaises(IndexError):
            g = forcefields.square_well([1.0, 1.0])

if __name__ == '__main__':
    unittest.main(exit=False)



import numpy as np


class lennard_jones_sigma_epsilon(object):
    r"""Calculate the energy or force for a pair of particles using the
    Lennard-Jones (sigma/epsilon variant) forcefield.

    Parameters
    ----------
    constants: float, array_like
        An array of length two consisting of the sigma (a) and epsilon (e)
        parameters for the 12-6 Lennard-Jones function

    """
    def __init__(self, constants):
        if len(constants) != 2:
            raise IndexError(f'There should be two constants per set, not {len(constants)}')
        
        self.sigma = constants[0]
        self.epsilon = constants[1]
        self.point_size = 1.3e10 * (self.sigma*(2**(1/6)))
    
    def energy(self, dr):
        r"""Calculate the energy for a pair of particles using the
        Lennard-Jones (sigma/epsilon variant) forcefield.

        .. math::
            E = \frac{4e*a^{12}}{dr^{12}} - \frac{4e*a^{6}}{dr^6}

        Attributes:
        ----------
        dr (float): The distance between particles.
 
        Returns
        -------
        float: array_like
        The potential energy between the particles.
        """
        self.energy = 4 * self.epsilon * np.power(self.sigma, 12) * np.power(dr, -12) - (
                        4 * self.epsilon * np.power(self.sigma, 6) * np.power(dr, -6)) 
        return self.energy 
    
    def force(self, dr):
        r"""Calculate the force for a pair of particles using the
        Lennard-Jones (sigma/epsilon variant) forcefield.

        .. math::
            f = \frac{48e*a^{12}}{dr^{13}} - \frac{24e*a^{6}}{dr^7}

        Attributes:
        ----------
        dr (float): The distance between particles.
 
        Returns
        -------
        float: array_like
        The force between the particles.
        """
        self.force = 48 * self.epsilon * np.power(self.sigma, 12) * np.power(
            dr, -13) - (24 * self.epsilon * np.power(self.sigma, 6) * np.power(dr, -7))
        return self.force
    
    def mixing(self, constants_2):
        r""" Calculates mixing for two sets of constants
        
        ..math::
            \sigma_{12} = \frac{\sigma_1 + \sigma_2}{2}
            \epsilon{12} = \sqrt{\epsilon_1 * \epsilon_2}
        
        Parameters:
        ----------
        constants_2: float, array_like
            The second set of constantss
        """
        sigma2 = constants_2[0]
        epsilon2 = constants_2[1]
        self.sigma = (self.sigma+sigma2)/2
        self.epsilon = np.sqrt(self.epsilon * epsilon2)


class lennard_jones(lennard_jones_sigma_epsilon):
    r"""Converts a/b variant values to sigma/epsilon variant
    then maps to lennard_jones_sigma_epsilon class

    ..math::
        \sigma = \frac{a}{b}^(\frac{1}{6})
        \sigma = \frace{b^2}{4*a}

    Parameters
    ----------
    constants: float, array_like
        An array of length two consisting of the A and B
        parameters for the 12-6 Lennard-Jones function
    """   
    def __init__(self, constants):
        if len(constants) != 2:
            raise IndexError(f'There should be two constants per set, not {len(constants)}')
        self.a = constants[0]
        self.b = constants[1]
        sigma = (self.a / self.b)**(1/6)
        epsilon = (self.b**2)/(4*self.a)
        super().__init__([sigma, epsilon])

    def mixing(self, constants_2):
        r"""Converts second set of a/b constants into sigma/epsilon
        for use in mixing method. Then converts changed self sigma/epsilon
        values back to a/b

        ..math::
            a = 4*\epsilon*(\sigma^12)
            b = 4*\epsilon*(\sigma^6)

        Parameters:
        ----------
        constants_2: float, array_like
            The second set of constantss
        """
        a2 = constants_2[0]
        b2 = constants_2[1]
        sigma2 = (a2 / b2)**(1/6)
        epsilon2 = (b2**2)/(4*a2)
        super().mixing([sigma2,epsilon2])
        self.a = 4 * self.epsilon * (self.sigma**12)
        self.b = 4 * self.epsilon * (self.sigma**6)


class buckingham(object):
    r""" Calculate the energy or force for a pair of particles using the
    Buckingham forcefield.

    Parameters
    ----------
    constants: float, array_like
        An array of length three consisting of the A, B and C parameters for
        the Buckingham function.

    """
    def __init__(self, constants):
        if len(constants) != 3:
            raise IndexError(f'There should be three constants per set, not {len(constants)}')
        self.a = constants[0]
        self.b = constants[1]
        self.c = constants[2]
        self.point_size = 8 # Needs better solution relevant to constants
    
    def energy(self, dr):
        r"""Calculate the energy for a pair of particles using the
        Buckingham forcefield.

        .. math::
            E = Ae^{(-Bdr)} - \frac{C}{dr^6}

        Attributes:
        ----------
        dr (float): The distance between particles.
 
        Returns
        -------
        float: array_like
        The potential energy between the particles.
        """
        energy = self.a * np.exp(- np.multiply(self.b, dr)) - self.c / np.power(dr, 6)
        # Cut out infinite values where r = 0
        if type(dr) != float:
            energy = np.array(energy)
            energy[np.where(energy > 10e300)] = 0
            energy[np.where(energy < -10e300)] = 0
        self.energy = energy
        return self.energy
    
    def force(self, dr):
        r"""Calculate the force for a pair of particles using the
        Buckingham forcefield.

        .. math::
            f = ABe^{(-Bdr)} - \frac{6C}{dr^7}

        Attributes:
        ----------
        dr (float): The distance between particles.
 
        Returns
        -------
        float: array_like
        The force between the particles.
        """
        force = self.a * self.b * np.exp(- np.multiply(self.b, dr)) - 6 * self.c / np.power(dr, 7)
        # Cut out infinite values where r = 0
        if type(dr) != float:
            force = np.array(force)
            force[np.where(force > 10e300)] = 0
            force[np.where(force < -10e300)] = 0
        self.force = force
        return self.force

    def mixing(self, constants2):
        r""" Calculates mixing for two sets of constants
        
        ..math::
            a_{12} = \sqrt{a_1 * a_2}
            b_{12} = \sqrt{b_1 * b_2}
            c_{12} = \sqrt{c_1 * c_2}
        
        Attributes:
        ----------
        constants_2: float, array_like
            The second set of constantss
        """
        self.a = np.sqrt(self.a*constants2[0])
        self.b = np.sqrt(self.b*constants2[1])
        self.c = np.sqrt(self.c*constants2[2])


class square_well(object):
    r'''Calculate the energy or force for a pair of particles using a
    square well model.

    Parameters
    ----------
    constants: float, array_like
        An array of length three consisting of the epsilon, sigma, and lambda
        parameters for the square well model.
    max_val: int (optional)
        Upper bound for values in square well - replaces usual infinite values
        
    '''
    def __init__(self, constants, max_val=np.inf):
        if len(constants) != 3:
            raise IndexError(f'There should be three constants per set, not {len(constants)}')
        self.epsilon = constants[0]
        self.sigma = constants[1]
        self.lamda = constants[2] #Spelling as lamda not lambda to avoid calling python lambda function
        self.max_val = max_val
        self.point_size = 10

    def energy(self, dr):
        r'''Calculate the energy for a pair of particles using a
        square well model.

        .. math::
            E = {
            if dr < sigma:
                E = max_val
            elif sigma <= dr < lambda * sigma:
                E = -epsilon
            elif r >= lambda * sigma:
                E = 0
            }

        Attributes:
        ----------
        dr (float): The distance between particles.
 
        Returns
        -------
        float: array_like
            The potential energy between the particles.
        '''

        if not isinstance(dr, np.ndarray):
            if isinstance(dr, list):
                dr = np.array(dr, dtype='float')
            elif isinstance(dr, float):
                dr = np.array([dr], dtype='float')

        E = np.zeros_like(dr)
        E = np.zeros_like(dr)
        E[np.where(dr < self.epsilon)] = self.max_val
        E[np.where(dr >= self.lamda * self.sigma)] = 0

        # apply mask for sigma <= dr < lambda * sigma
        a = self.sigma <= dr
        b = dr < self.lamda * self.sigma
        E[np.where(a & b)] = -self.epsilon

        if len(E) == 1:
            self.energy = float(E[0])
        else:
            self.energy = np.array(E, dtype='float')
            
        return self.energy
        
    def force(self):
        r'''The force of a pair of particles using a square well model is given by:

        .. math::
        f = {
        if sigma <= dr < lambda * sigma:
            f = inf
        else:
            f = 0
        }

        Therefore the force here will always be infinite, and therefore not possible to simulate
        '''
        raise ValueError("Force is infinite at sigma <= dr < lambda * sigma")


import numpy as np
from pylj import pairwise as heavy
from pylj import forcefields as ff


def initialise(
    number_of_particles,
    temperature,
    box_length,
    init_conf,
    timestep_length=1e-14,
    mass=39.948,
    constants=[[1.363e-134, 9.273e-78]],
    forcefield=ff.lennard_jones
):
    """Initialise the particle positions (this can be either as a square or
    random arrangement), velocities (based on the temperature defined, and
    calculate the initial forces/accelerations.

    Parameters
    ----------
    number_of_particles: int
        Number of particles to simulate.
    temperature: float
        Initial temperature of the particles, in Kelvin.
    box_length: float
        Length of a single dimension of the simulation square, in Angstrom.
    init_conf: string, optional
        The way that the particles are initially positioned. Should be one of:
        - 'square'
        - 'random'
    timestep_length: float (optional)
        Length for each Velocity-Verlet integration step, in seconds.
    mass: float (optional)
        The mass of the particles being simulated.
    constants: float, array_like (optional)
        The values of the constants for the forcefield used.
    forcefield: function (optional)
        The particular forcefield to be used to find the energy and forces.

    Returns
    -------
    System
        System information.
    """
    from pylj import util

    system = util.System(
        number_of_particles,
        temperature,
        box_length,
        constants,
        forcefield,
        mass,
        init_conf=init_conf,
        timestep_length=timestep_length
    )
    v = np.random.rand(system.particles.size, 2, 12)
    v = np.sum(v, axis=2) - 6.0
    mass_kg = mass * 1.6605e-27
    v = v * np.sqrt(1.3806e-23 * system.init_temp / mass_kg)
    v = v - np.average(v)
    system.particles["xvelocity"] = v[:, 0]
    system.particles["yvelocity"] = v[:, 1]
    return system


def initialize(
    number_particles, temperature, box_length, init_conf, timestep_length=1e-14
):
    """Maps to the md.initialise function to account for US english spelling.
    """
    a = initialise(
        number_particles, temperature, box_length, init_conf, timestep_length
    )
    return a


#Jit tag here had to be removed
def velocity_verlet(
    particles, timestep_length, box_length, cut_off, constants, forcefield, mass
):
    """Uses the Velocity-Verlet integrator to move forward in time. The
    Updates the particles positions and velocities in terms of the Velocity
    Verlet algorithm. Also calculates the instanteous temperature, pressure,
    and force and appends these to the appropriate system array.
    Parameters
    ----------
    particles: util.particle_dt, array_like
        Information about the particles.
    timestep_length: float
        Length for each Velocity-Verlet integration step, in seconds.
    box_length: float
        Length of a single dimension of the simulation square, in Angstrom.
    Returns
    -------
    util.particle_dt, array_like:
        Information about the particles, with new positions and velocities.
    """
    xposition_store = list(particles["xposition"])
    yposition_store = list(particles["yposition"])
    pos, prev_pos = update_positions(
        [particles["xposition"], particles["yposition"]],
        [particles["xprevious_position"], particles["yprevious_position"]],
        [particles["xvelocity"], particles["yvelocity"]],
        [particles["xacceleration"], particles["yacceleration"]],
        timestep_length,
        box_length,
    )
    [particles["xposition"], particles["yposition"]] = pos
    [particles["xprevious_position"], particles["yprevious_position"]] = pos
    xacceleration_store = list(particles["xacceleration"])
    yacceleration_store = list(particles["yacceleration"])
    particles, distances, forces, energies = heavy.compute_force(
        particles, box_length, cut_off, constants, forcefield, mass
    )
    [particles["xvelocity"], particles["yvelocity"]] = update_velocities(
        [particles["xvelocity"], particles["yvelocity"]],
        [xacceleration_store, yacceleration_store],
        [particles["xacceleration"], particles["yacceleration"]],
        timestep_length,
    )
    particles["xprevious_position"] = xposition_store
    particles["yprevious_position"] = yposition_store
    return particles, distances, forces, energies


def sample(particles, box_length, initial_particles, system):
    """Sample parameters of interest in the simulation.
    Parameters
    ----------
    particles: util.particle_dt, array_like
        Information about the particles.
    box_length: float
        Length of a single dimension of the simulation square, in Angstrom.
    initial_particles: util.particle_dt, array-like
        Information about the initial particle conformation.
    system: System
        Details about the whole system
    Returns
    -------
    System:
        Details about the whole system, with the new temperature, pressure,
        msd, and force appended to the appropriate
        arrays.
    """
    temperature_new = calculate_temperature(particles, system.mass)
    system.temperature_sample = np.append(system.temperature_sample, temperature_new)
    pressure_new = heavy.calculate_pressure(
        particles,
        box_length,
        temperature_new,
        system.cut_off,
        system.constants,
        system.forcefield,
        system.mass
    )
    msd_new = calculate_msd(particles, initial_particles, box_length)
    system.pressure_sample = np.append(system.pressure_sample, pressure_new)
    system.force_sample = np.append(system.force_sample, np.sum(system.forces))
    system.energy_sample = np.append(system.energy_sample, np.sum(system.energies))
    system.msd_sample = np.append(system.msd_sample, msd_new)
    return system


def calculate_msd(particles, initial_particles, box_length):
    """Determines the mean squared displacement of the particles in the system.
    Parameters
    ----------
    particles: util.particle_dt, array_like
        Information about the particles.
    initial_particles: util.particle_dt, array_like
        Information about the initial state of the particles.
    box_length: float
        Size of the cell vector.
    Returns
    -------
    float:
        Mean squared deviation for the particles at the given timestep.
    """
    xpos = np.array(particles["xposition"])
    ypos = np.array(particles["yposition"])
    dxinst = xpos - particles["xprevious_position"]
    dyinst = ypos - particles["yprevious_position"]
    for i in range(0, particles["xposition"].size):
        if np.abs(dxinst[i]) > 0.5 * box_length:
            if xpos[i] <= 0.5 * box_length:
                particles["xpbccount"][i] += 1
            if xpos[i] > 0.5 * box_length:
                particles["xpbccount"][i] -= 1
        xpos[i] += box_length * particles["xpbccount"][i]
        if np.abs(dyinst[i]) > 0.5 * box_length:
            if ypos[i] <= 0.5 * box_length:
                particles["ypbccount"][i] += 1
            if ypos[i] > 0.5 * box_length:
                particles["ypbccount"][i] -= 1
        ypos[i] += box_length * particles["ypbccount"][i]
    dx = xpos - initial_particles["xposition"]
    dy = ypos - initial_particles["yposition"]
    dr = np.sqrt(dx * dx + dy * dy)
    return np.average(dr ** 2)


def update_positions(
    positions, old_positions, velocities, accelerations, timestep_length, box_length
):
    """Update the particle positions using the Velocity-Verlet integrator.
    Parameters
    ----------
    positions: (2, N) array_like
        Where N is the number of particles, and the first row are the x
        positions and the second row the y positions.
    old_positions: (2, N) array_like
        Where N is the number of particles, and the first row are the
        previous x positions and the second row are the y positions.
    velocities: (2, N) array_like
        Where N is the number of particles, and the first row are the x
        velocities and the second row the y velocities.
    accelerations: (2, N) array_like
        Where N is the number of particles, and the first row are the x
        accelerations and the second row the y
        accelerations.
    timestep_length: float
        Length for each Velocity-Verlet integration step, in seconds.
    box_length: float
        Length of a single dimension of the simulation square, in Angstrom.
    Returns
    -------
    (2, N) array_like:
        Updated positions.
    """
    old_positions[0] = np.array(positions[0])
    old_positions[1] = np.array(positions[1])
    positions[0] += (velocities[0] * timestep_length) + (
        0.5 * accelerations[0] * timestep_length * timestep_length
    )
    positions[1] += (velocities[1] * timestep_length) + (
        0.5 * accelerations[1] * timestep_length * timestep_length
    )
    positions[0] = positions[0] % box_length
    positions[1] = positions[1] % box_length
    return [positions[0], positions[1]], [old_positions[0], old_positions[1]]


def update_velocities(
    velocities, accelerations_old, accelerations_new, timestep_length
):
    """Update the particle velocities using the Velocity-Verlet algoritm.
    Parameters
    ----------
    velocities: (2, N) array_like
        Where N is the number of particles, and the first row are the x
        velocities and the second row the y velocities.
    accelerations: (2, N) array_like
        Where N is the number of particles, and the first row are the x
        accelerations and the second row the y
        accelerations.
    timestep_length: float
        Length for each Velocity-Verlet integration step, in seconds.
    Returns
    -------
    (2, N) array_like:
        Updated velocities.
    """
    velocities[0] += (
        0.5 * (accelerations_old[0] + accelerations_new[0]) * timestep_length
    )
    velocities[1] += (
        0.5 * (accelerations_old[1] + accelerations_new[1]) * timestep_length
    )
    return [velocities[0], velocities[1]]


def calculate_temperature(particles, mass):
    """Determine the instantaneous temperature of the system.
    Parameters
    ----------
    particles: util.particle_dt, array_like
        Information about the particles.
    Returns
    -------
    float:
        Calculated instantaneous simulation temperature.
    """
    boltzmann_constant = 1.3806e-23  # joules/kelvin
    atomic_mass_unit = 1.660539e-27  # kilograms
    mass_kg = mass * atomic_mass_unit  # kilograms
    v = np.sqrt(
        (particles["xvelocity"] * particles["xvelocity"])
        + (particles["yvelocity"] * particles["yvelocity"])
    )
    k = 0.5 * np.sum(mass_kg * v * v)
    t = k / (particles.size * boltzmann_constant)
    return t


def compute_force(particles, box_length, cut_off, constants, forcefield, mass):
    r"""Calculates the forces and therefore the accelerations on each of the
    particles in the simulation.
    Parameters
    ----------
    particles: util.particle_dt, array_like
        Information about the particles.
    box_length: float
        Length of a single dimension of the simulation square, in Angstrom.
    cut_off: float
        The distance greater than which the forces between particles is taken
        as zero.
    constants: float, array_like (optional)
        The constants associated with the particular forcefield used, e.g. for
        the function forcefields.lennard_jones, theses are [A, B]
    forcefield: function (optional)
        The particular forcefield to be used to find the energy and forces.
    mass: float (optional)
        The mass of the particle being simulated (units of atomic mass units).
    Returns
    -------
    util.particle_dt, array_like
        Information about particles, with updated accelerations and forces.
    float, array_like
        Current distances between pairs of particles in the simulation.
    float, array_like
        Current forces between pairs of particles in the simulation.
    float, array_like
        Current energies between pairs of particles in the simulation.
    """
    part, dist, forces, energies = heavy.compute_force(
        particles, box_length, cut_off, constants, forcefield, mass=mass
    )
    return part, dist, forces, energies


def heat_bath(particles, temperature_sample, bath_temperature):
    r"""Rescales the velocities of the particles in the system to control the
    temperature of the simulation. Thereby allowing for an NVT ensemble. The
    velocities are rescaled according the following relationship,
    .. math::
        v_{\text{new}} = v_{\text{old}} \times
        \sqrt{\frac{T_{\text{desired}}}{\bar{T}}}
    Parameters
    ----------
    particles: util.particle_dt, array_like
        Information about the particles.
    temperature_sample: float, array_like
        The temperature at each timestep in the simulation.
    bath_temp: float
        The desired temperature of the simulation.
    Returns
    -------
    util.particle_dt, array_like
        Information about the particles with new, rescaled velocities.
    """
    particles = heavy.heat_bath(particles, temperature_sample, bath_temperature)
    return particles



#! /usr/bin/env python
# System imports
from setuptools import setup, Extension, find_packages
from os import path
import io

packages = find_packages()

# versioning
MAJOR = 1
MINOR = 5
MICRO = 2
ISRELEASED = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)


this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

info = {
        'name': 'pylj',
        'description': 'Simple teaching tool for classical MD simulation',
        'author': 'Andrew R. McCluskey',
        'author_email': 'arm61@bath.ac.uk',
        'packages': packages,
        'include_package_data': True,
        'setup_requires': ['jupyter', 'numpy', 'matplotlib', 'cython'],
        'install_requires': ['jupyter', 'numpy', 'matplotlib', 'cython'],
        'version': VERSION,
        'license': 'MIT',
        'long_description': long_description,
        'long_description_content_type': 'text/markdown',
        'classifiers': ['Development Status :: 5 - Production/Stable',
                        'Framework :: Jupyter',
                        'Intended Audience :: Science/Research',
                        'License :: OSI Approved :: MIT License',
                        'Natural Language :: English',
                        'Operating System :: OS Independent',
                        'Programming Language :: Python :: 2.7',
                        'Programming Language :: Python :: 3.5',
                        'Programming Language :: Python :: 3.6',
                        'Programming Language :: Python :: 3.7',
                        'Topic :: Scientific/Engineering',
                        'Topic :: Scientific/Engineering :: Chemistry',
                        'Topic :: Scientific/Engineering :: Physics']
        }

####################################################################
# this is where setup starts
####################################################################


def setup_package():
    setup(**info)


if __name__ == '__main__':
    setup_package()



from __future__ import division
import numpy as np
import webbrowser
from pylj import md, mc


class System:
    """Simulation system.
    This class is designed to store all of the information about the job that
    is being run. This includes the particles
    object, as will as sampling objects such as the temperature, pressure, etc.
    arrays.
    Parameters
    ----------
    number_of_particles: int
        Number of particles to simulate.
    temperature: float
        Initial temperature of the particles, in Kelvin.
    box_length: float
        Length of a single dimension of the simulation square, in Angstrom.
    init_conf: string, optional
        The way that the particles are initially positioned. Should be one of:
        - 'square'
        - 'random'
    timestep_length: float (optional)
        Length for each Velocity-Verlet integration step, in seconds.
    cut_off: float (optional)
        The distance apart that the particles must be to consider there
        interaction to be negliable.
    constants: float, array_like (optional)
        The values of the constants for the forcefield used.
    mass: float (optional)
        The mass of the particles being simulated.
    forcefield: function (optional)
        The particular forcefield to be used to find the energy and forces.
    """

    def __init__(
        self,
        number_of_particles,
        temperature,
        box_length,
        constants,
        forcefield,
        mass,
        init_conf="square",
        timestep_length=1e-14,
        cut_off=15
    ):
        self.number_of_particles = number_of_particles
        self.init_temp = temperature
        self.constants = constants
        self.mass = mass
        self.forcefield = forcefield
        self.type_identifiers = None
        self.particle_list = None
        self.long_const = None
        self.types = None
        self.point_sizes = None
        self.setup_point_sizes()
        self.setup_type_identifiers()
        self.setup_types()
        if box_length <= 600:
            self.box_length = box_length * 1e-10
        else:
            raise AttributeError(
                "With a box length of {} the particles are "
                "probably too small to be seen in the "
                "viewer. Try something (much) less than "
                "600.".format(box_length)
            )
        if box_length >= 4:
            self.box_length = box_length * 1e-10
        else:
            raise AttributeError(
                "With a box length of {} the cell is too "
                "small to really hold more than one "
                "particle.".format(box_length)
            )
        self.timestep_length = timestep_length
        self.particles = None
        if init_conf == "square":
            self.square()
        elif init_conf == "random":
            self.random()
        else:
            raise NotImplementedError(
                "The initial configuration type {} is "
                "not recognised. Available options are: "
                "square or random".format(init_conf)
            )
        if box_length > 30:
            self.cut_off = cut_off * 1e-10
        else:
            self.cut_off = box_length / 2 * 1e-10
        self.step = 0
        self.time = 0.0
        self.distances = np.zeros(self.number_of_pairs())
        self.forces = np.zeros(self.number_of_pairs())
        self.energies = np.zeros(self.number_of_pairs())
        self.temperature_sample = np.array([])
        self.pressure_sample = np.array([])
        self.force_sample = np.array([])
        self.msd_sample = np.array([])
        self.energy_sample = np.array([])
        self.initial_particles = np.array(self.particles)
        self.position_store = [0, 0]
        self.old_energy = 0
        self.new_energy = 0
        self.random_particle = 0

    def number_of_pairs(self):
        """Calculates the number of pairwise interactions in the simulation.
        Returns
        -------
        int:
            Number of pairwise interactions in the system.
        """
        return int((self.number_of_particles - 1) * self.number_of_particles / 2)

    def square(self):
        """Sets the initial positions of the particles on a square lattice.
        """
        part_dt = particle_dt()
        self.particles = np.zeros(self.number_of_particles, dtype=part_dt)
        self.particles['types'] = self.types
        m = int(np.ceil(np.sqrt(self.number_of_particles)))
        d = self.box_length / m
        n = 0
        for i in range(0, m):
            for j in range(0, m):
                if n < self.number_of_particles:
                    self.particles[n]["xposition"] = (i + 0.5) * d
                    self.particles[n]["yposition"] = (j + 0.5) * d
                    n += 1

    def random(self):
        """Sets the initial positions of the particles in a random arrangement.
        """
        part_dt = particle_dt()
        self.particles = np.zeros(self.number_of_particles, dtype=part_dt)
        self.particles['types'] = self.types
        num_part = self.number_of_particles
        self.particles["xposition"] = np.random.uniform(0, self.box_length, num_part)
        self.particles["yposition"] = np.random.uniform(0, self.box_length, num_part)

    def setup_types(self):
        """Sets the long constants and types arrays of the particles
        """
        long_const = []
        types = []
        particle_list = []
        for i in range(self.number_of_particles):
            # Get set of constants and index
            constants_type = self.constants[i%len(self.constants)]
            particle_type = f'{i%len(self.constants)}'
            # Append to lists
            long_const.append(constants_type)
            types.append(particle_type)
            particle = Particle(constants_type, i, self.mass, particle_type)
            particle.add(particle_list)
        self.particle_list = particle_list
        self.long_const = long_const
        self.types = types

    def setup_type_identifiers(self):
        """Sets type-identifers arrays - legacy method now only used for plotting
        """
        # Creates arrays to identify which particle is in which type
        number_of_types = len(self.constants)
        type_identifiers = np.zeros((number_of_types,self.number_of_particles))
        i = 0
        while i < self.number_of_particles:
            for k in range(number_of_types):
                if i < self.number_of_particles:
                    type_identifiers[k][i] = 1
                    i+=1
        self.type_identifiers = type_identifiers

    def setup_point_sizes(self):
        """Sets point sizes for use in plotting
        """
        point_sizes = []
        for pair in self.constants:
            size = self.forcefield(pair).point_size
            point_sizes.append(size)
        self.point_sizes = point_sizes

    def compute_force(self):
        """Maps to the compute_force function in either the comp (if Cython is
        installed) or the pairwise module and allows for a cleaner interface.
        """
        part, dist, forces, energies = md.compute_force(
            self.particles,
            self.box_length,
            self.cut_off,
            self.constants,
            self.forcefield,
            self.mass,
        )
        self.particles = part
        self.distances = dist
        self.forces = forces
        self.energies = energies

    def compute_energy(self):
        """Maps to the compute_force function, as this also calculates energy
        """
        self.compute_force()

    #Jit tag here had to be removed
    def integrate(self, method):
        """Maps the chosen integration method.
        Parameters
        ----------
        method: method
            The integration method to be used, e.g. md.velocity_verlet.
        """
        self.particles, self.distances, self.forces, self.energies = method(
            self.particles,
            self.timestep_length,
            self.box_length,
            self.cut_off,
            self.constants,
            self.forcefield,
            self.mass
        )

    def md_sample(self):
        """Maps to the md.sample function.
        """
        self = md.sample(self.particles, self.box_length, self.initial_particles, self)

    def heat_bath(self, bath_temperature):
        """Maps to the heat_bath function in either the comp (if Cython is
        installed) or the pairwise modules.
        Parameters
        ----------
        target_temperature: float
            The target temperature for the simulation.
        """
        self.particles = md.heat_bath(
            self.particles, self.temperature_sample, bath_temperature
        )

    def mc_sample(self):
        """Maps to the mc.sample function.
        Parameters
        ----------
        energy: float
            Energy to add to the sample
        """
        mc.sample(self.old_energy, self)

    def select_random_particle(self):
        """Maps to the mc.select_random_particle function.
        """
        self.random_particle, self.position_store = mc.select_random_particle(
            self.particles
        )

    def new_random_position(self):
        """Maps to the mc.get_new_particle function.
        """
        self.particles = mc.get_new_particle(
            self.particles, self.random_particle, self.box_length
        )

    def accept(self):
        """Maps to the mc.accept function.
        """
        self.old_energy = mc.accept(self.new_energy)

    def reject(self):
        """Maps to the mc.reject function.
        """
        self.particles = mc.reject(
            self.position_store, self.particles, self.random_particle
        )

class Particle:
    def __init__(self,
               constants,
               index,
               mass,
               particle_type
               ):
        self.constants = constants
        self.index = index
        self.mass = mass
        self.type = particle_type

    def add(self, particles):
        particles.append(self)
        return particles

def __cite__():  # pragma: no cover
    """This function will launch the website for the JOSE publication on
    pylj."""
    webbrowser.open("http://jose.theoj.org/papers/58daa1a1a564dc8e0f99ffcdae20eb1d")


def __version__():  # pragma: no cover
    """This will print the number of the pylj version currently in use."""
    major = 1
    minor = 4
    micro = 1
    print("pylj-{:d}.{:d}.{:d}".format(major, minor, micro))


def particle_dt():
    """Builds the data type for the particles, this consists of:
    - xposition and yposition
    - xvelocity and yvelocity
    - xacceleration and yacceleration
    - xprevious_position and yprevious_position
    - xforce and yforce
    - energy
    - types
    """
    return np.dtype(
        [
            ("xposition", np.float64),
            ("yposition", np.float64),
            ("xvelocity", np.float64),
            ("yvelocity", np.float64),
            ("xacceleration", np.float64),
            ("yacceleration", np.float64),
            ("xprevious_position", np.float64),
            ("yprevious_position", np.float64),
            ("energy", np.float64),
            ("xpbccount", int),
            ("ypbccount", int),
            ("types", list)
        ]
    )
